﻿/*
 * Copyright 2021 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

using System;
using System.Collections.Generic;

using PluginCommon;

namespace Apple.ProSel8 {
    /// <summary>
    /// Format inline strings in ProSel-8 programs (e.g. Cat.Doctor), and SmartPort DISPATCH
    /// calls.
    ///
    /// The strings are null-terminated high ASCII, with embedded HTAB values (low bit clear).
    /// </summary>
    public class FmtInline : MarshalByRefObject, IPlugin, IPlugin_SymbolList, IPlugin_InlineJsr {
        private IApplication mAppRef;
        private byte[] mFileData;

        private class NameMap {
            public string Prefix { get; private set; }
            public InlineKind Kind { get; private set; }
            public NameMap(string prefix, InlineKind kind) {
                Prefix = prefix;
                Kind = kind;
            }
        };
        private enum InlineKind { Unknown = 0, InHZ, InSP };
        private static NameMap[] sMap = {
            new NameMap("InHZ_", InlineKind.InHZ),
            new NameMap("InSP_", InlineKind.InSP),
        };

        // Map of target address to inline data format.
        private Dictionary<int, InlineKind> mInlineAddrs = new Dictionary<int, InlineKind>();

        // IPlugin
        public string Identifier {
            get { return "ProSel-8 inline data formatter"; }
        }

        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;

            mAppRef.DebugLog("ProSel8(id=" + AppDomain.CurrentDomain.Id + "): prepare()");
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        // IPlugin_SymbolList
        public void UpdateSymbolList(List<PlSymbol> plSyms) {
            mInlineAddrs.Clear();

            // Find matching symbols.
            foreach (PlSymbol sym in plSyms) {
                if (sym.Value == AddressTranslate.NON_ADDR) {
                    continue;
                }
                foreach (NameMap map in sMap) {
                    if (sym.Label.StartsWith(map.Prefix)) {
                        // Symbol values (addresses) may not be unique.
                        mInlineAddrs[sym.Value] = map.Kind;
                        break;
                    }
                }
            }
            mAppRef.DebugLog("Found matches for " + mInlineAddrs.Count + " addresses");
        }

        // IPlugin_SymbolList
        public bool IsLabelSignificant(string beforeLabel, string afterLabel) {
            return DoesLabelMatch(beforeLabel) || DoesLabelMatch(afterLabel);
        }
        private static bool DoesLabelMatch(string label) {
            foreach (NameMap map in sMap) {
                if (label.StartsWith(map.Prefix)) {
                    return true;
                }
            }
            return false;
        }

        // IPlugin_InlineJsr
        public void CheckJsr(int offset, int operand, out bool noContinue) {
            noContinue = false;

            // See if the address matches something we know about.
            InlineKind kind;
            if (!mInlineAddrs.TryGetValue(operand, out kind)) {
                // JSR destination address not special.
                return;
            }

            offset += 3;    // move past JSR

            switch (kind) {
                case InlineKind.InHZ:
                    // Null-terminated ASCII string.
                    FormatNullTermString(offset);
                    break;
                case InlineKind.InSP:
                    // SmartPort DISPATCH call.
                    FormatSmartPortDispatch(offset);
                    break;
            }
        }

        private void FormatNullTermString(int offset) {
            if (offset < 0 || offset >= mFileData.Length) {
                return;     // first byte is not inside file
            }

            // search for the terminating null byte
            int nullOff = offset;
            while (nullOff < mFileData.Length) {
                if (mFileData[nullOff] == 0x00) {
                    break;
                }
                nullOff++;
            }
            if (nullOff == mFileData.Length) {
                mAppRef.DebugLog("Unable to find end of null-terminated string at +" +
                    offset.ToString("x6"));
                return;
            }

            DataSubType stype = DataSubType.HighAscii;

            mAppRef.SetInlineDataFormat(offset, nullOff - offset + 1,
                DataType.StringNullTerm, stype, null);
        }

        private void FormatSmartPortDispatch(int offset) {
            if (offset < 0 || offset + 3 >= mFileData.Length) {
                return;
            }

            mAppRef.SetInlineDataFormat(offset, 1, DataType.NumericLE, DataSubType.Hex, null);
            mAppRef.SetInlineDataFormat(offset+1, 2, DataType.NumericLE, DataSubType.Address, null);
        }
    }
}
